from __future__ import with_statement
from __future__ import absolute_import
from __future__ import division


import math

from shadowfin import ShadowFin
from vector2d import Vector2D as vec2
import vector2d
from OpenGL.GL import *


class ConvexHull():
    def __init__(self, pos, points, depth=0.0, color=(1.0, 0.0, 0.0, 0.0)):
        new_points = []
        for point in points:
            new_points.append((point[0] + pos[0], point[1] + pos[1]))
        self.points = new_points

        self.position = pos
        self.depth = depth
        self.color = color

    def draw_shadow_geometry(self, light):
        # Calculate all the front facing sides
        first, last = None, None
        self.last_node_frontfacing = None
        for x in xrange(-1, len(self.points), 1):
            current_point = self.points[x]
            prev_point = self.points[x - 1]

            nv = vec2(current_point[1] - prev_point[1],
                    current_point[0] - prev_point[0])

            lv = vec2(current_point[0] - light.position[0],
                    current_point[1] - light.position[1])

            # Check if the face is front-facing
            if (nv[0] * -1 * lv[0]) + (nv[1] * lv[1]) > 0:
                if self.last_node_frontfacing == False:
                    last = self.points.index(prev_point)
                self.last_node_frontfacing = True
            else:
                if self.last_node_frontfacing:
                    first = self.points.index(prev_point)
                self.last_node_frontfacing = False

        if first is None or last is None:
            # The light source is inside the object
            return

        # Create shadow fins
        start_fins, first, first_vector = self.create_shadowfins(light, first, 1)
        end_fins, last, last_vector = self.create_shadowfins(light, last, -1)

        # Render shadow fins
        for fin in start_fins + end_fins:
            fin.render()

        # Get a list of all the back edges
        backpoints = []
        for x in xrange(first, first + len(self.points), 1):
            backpoints.insert(0, self.points[(x)%len(self.points)])
            if x%len(self.points) == last:
                break

        # Figure out the length of the back edges. We'll use this later for
        # weighted average between the shadow fins to find our umbra vectors.
        back_length = [0]
        sum_back_length = 0
        for x in xrange(1, len(backpoints), 1):
            l = vector2d.from_points(backpoints[x - 1], backpoints[x]).length()
            back_length.insert(0, l)
            sum_back_length += l


        # Draw the shadow geometry using a triangle strip
        glColor4f(0.0, 0.0, 0.0, 0.0)
        glBegin(GL_TRIANGLE_STRIP)
        a = 0
        for x in xrange(len(backpoints)):
            point = backpoints[x]
            glVertex3f(point[0], point[1], self.depth)
            # Draw our umbra using weighted average vectors
            if x != len(backpoints) - 2:
                glVertex3f(point[0] + (first_vector[0] * (a / sum_back_length)) + (last_vector[0] * (1 - (a / sum_back_length))),
                        point[1] + (first_vector[1] * (a / sum_back_length)) + (last_vector[1] * (1 - (a / sum_back_length))), self.depth)
            else:
                glVertex3f(point[0] + first_vector[0], point[1] + first_vector[1], self.depth)
            a += back_length[x]
        glEnd()


    def create_shadowfins(self, light, origin, step):
        shadowfins = []

        # Go backwards to see if we need any shadow fins
        i = origin
        while True:
            p1 = self.points[i]

            # Make sure we wrap around
            i -= step
            if i < 0:
                i = len(self.points) - 1
            elif i == len(self.points):
                i = 0

            p0 = self.points[i]

            edge = vector2d.from_points(p1, p0)
            edge.normalize()

            shadowfin = ShadowFin(p0)
            shadowfin.index = i

            angle = edge.angle() - light.outer_vector(p0, step).angle()

            if step == 1:
                if angle < 0 or angle > math.pi * 0.5:
                    break
            elif step == -1:
                # Make sure the angle is within the right quadrant
                if angle > math.pi:
                    angle -= math.pi * 2
                if angle > 0 or angle < -math.pi * 0.5:
                    break

            shadowfin.set_outer(light.outer_vector(p0, step))
            shadowfin.set_inner(edge * light.inner_vector(p0, step).length())

            shadowfins.append(shadowfin)
            #break

        # Go forwards and see if we need any shadow fins
        i = origin
        while True:
            shadowfin = ShadowFin(self.points[i])
            shadowfin.index = i

            shadowfin.set_outer(light.outer_vector(self.points[i], step))
            shadowfin.set_inner(light.inner_vector(self.points[i], step))

            if len(shadowfins) > 0:
                shadowfin.set_outer(shadowfins[0].inner)

            p0 = self.points[i]

            # Make sure we wrap around
            i += step
            if i < 0:
                i = len(self.points) - 1
            elif i == len(self.points):
                i = 0

            p1 = self.points[i]

            edge = vector2d.from_points(p1, p0)
            edge.normalize()

            done = True
            penumbra = shadowfin.outer.copy()
            penumbra.normalize()
            umbra = shadowfin.inner.copy()
            umbra.normalize()
            if math.acos(vector2d.dot(edge, penumbra)) < \
                    math.acos(vector2d.dot(umbra, penumbra)):
                shadowfin.inner = edge * light.outer_vector(p0, step).length()
                done = False

            shadowfins.insert(0, shadowfin)

            if done:
                break

        # Get the total angle
        sum_angles = 0
        for x in xrange(len(shadowfins)):
            sum_angles += shadowfins[x].angle()

        # Calculate the inner and outer intensity of the shadowfins
        angle = 0
        for x in xrange(len(shadowfins)):
            shadowfins[x].set_umbra_intensity(angle / sum_angles)
            angle += shadowfins[x].angle()
            shadowfins[x].set_penumbra_intensity(angle / sum_angles)

        # We'll use these for our umbra generation
        return shadowfins, shadowfins[0].index, shadowfins[0].inner

    def render(self):
        glColor4f(self.color[0], self.color[1], self.color[2], 1.0)
        glBegin(GL_POLYGON)
        for point in self.points:
            glVertex3f(point[0], point[1], self.depth)
        glEnd()
