"""This class contains two ways to render to a texture, one using framebuffer
objects. The other using opengl 1.1 methods."""
from __future__ import with_statement
from __future__ import absolute_import
from __future__ import division


import math
import ctypes

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GL.EXT.framebuffer_object import *


class FrameBufferObject(object):
    def __init__(self, size=None):
        # See if we support non-p2 textures
        try:
            self.tex_target = GL_TEXTURE_RECTANGLE_EXT
            support_nonp2 = True
        except:
            try:
                self.tex_target = GL_TEXTURE_RECTANGLE_ARB
                support_nonp2 = True
            except:
                try:
                    self.tex_target = GL_TEXTURE_RECTANGLE_NV
                    support_nonp2 = True
                except:
                    support_nonp2 = False
                    self.tex_target = GL_TEXTURE_2D
        # FIXME: This should be handled properly
        self.tex_target = GL_TEXTURE_2D
        if size:
            # Actual size to draw
            self.width, self.height = size
            # Internal size
            if not support_nonp2:
                self.a_width, self.a_height = self.next_poweroftwo(size[0]), \
                    self.next_poweroftwo(size[1])
            else:
                self.a_width, self.a_height = size
        else:
            x, y, width, height = glGetIntegerv(GL_VIEWPORT)
            # Actual size to draw
            self.width, self.height = width, height
            # Internal size
            if not support_nonp2:
                self.a_width, self.a_height = self.next_poweroftwo(width), \
                    self.next_poweroftwo(height)
            else:
                self.a_width, self.a_height = self.width, self.height
        self.fbo = None
        self.depthbuffer = None
        self.tex = None
        self.enabled = False

        # Create the FBO right away
        self.create_fbo()

    def __del__(self):
        if self.fbo and self.depthbuffer:
            glDeleteFramebuffersEXT(1, self.fbo)
            glDeleteRenderbuffersEXT(1, self.depthbuffer)
        glDeleteTextures([self.tex])

    def next_poweroftwo(self, n):
        return int(pow(2, math.ceil(math.log(n) / math.log(2))))

    def create_fbo(self):
        # Generate the framebuffer
        if not glGenFramebuffersEXT:
            raise Exception("Framebuffer objects not supported")
        self.fbo = glGenFramebuffersEXT(1)
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self.fbo)

        # Create a depthbuffer
        self.depthbuffer = glGenRenderbuffersEXT(1)
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, self.depthbuffer)

        # Make depth buffer width, height in size
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT,
                self.a_width, self.a_height)
        # Attach depthbuffer to framebuffer
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, self.depthbuffer)

        # Create a texture to render to in size width, height
        self.tex = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.tex)
        glTexImage2D(self.tex_target, 0, GL_RGBA8, self.a_width, self.a_height,
                0, GL_RGBA, GL_UNSIGNED_BYTE, None)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                GL_LINEAR)


        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                GL_TEXTURE_2D, self.tex, 0)

        # Check that all went well
        status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT)
        print status
        if status != GL_FRAMEBUFFER_COMPLETE_EXT:
            if status == 36061:
                # FIXME: We need to handle this somehow
                raise Exception("Framebuffer objects not supported!")
            else:
                raise Exception("Could not create framebuffer!")
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)
        #glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0)

    def enable(self):
        if self.enabled:
            return
        if not self.fbo:
            self.create_fbo()
        self.enabled = True
        self.old_x, self.old_y, self.old_w, self.old_h = glGetIntegerv(GL_VIEWPORT)
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self.fbo)
        #glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, self.depthbuffer)
        
        glClearColor(0.0, 0.0, 0.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        glViewport(0, 0, self.a_width, self.a_height)

    def disable(self):
        if not self.enabled:
            return
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)
        #glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0)
        glViewport(self.old_x, self.old_y, self.old_w, self.old_h)
        self.enabled = False

    def render(self, alpha=1.0):
        if not self.enabled:
            glEnable(GL_TEXTURE_2D)
            glBindTexture(GL_TEXTURE_2D, self.tex)
            glColor4f(1.0, 1.0, 1.0, alpha)
            glBegin(GL_QUADS)
            glTexCoord2f(0.0, 0.0)
            glVertex2f(0.0, 0.0)
            glTexCoord2f(1.0, 0.0)
            glVertex2f(self.width, 0.0)
            glTexCoord2f(1.0, 1.0)
            glVertex2f(self.width, self.height)
            glTexCoord2f(0.0, 1.0)
            glVertex2f(0.0, self.height)
            glEnd()
            glBindTexture(GL_TEXTURE_2D, 0)
            glDisable(GL_TEXTURE_2D)

class BufferToTexture():
    def __init__(self, size=None, buffer=GL_BACK, do_clear_buffer=False):
        self.tex = None
        self.buffer = buffer
        self.do_clear_buffer = do_clear_buffer

    def next_poweroftwo(self, n):
        return int(pow(2, math.ceil(math.log(n) / math.log(2))))

    def create_texture(self):
        x, y, width, height = glGetIntegerv(GL_VIEWPORT)
        self.width = self.next_poweroftwo(width)
        self.height = self.next_poweroftwo(height)
        self.tex = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.tex)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

    def enable(self):
        if not self.tex:
            self.create_texture()

    def disable(self):
        glBindTexture(GL_TEXTURE_2D, self.tex)
        glReadBuffer(self.buffer)
        glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 0, 0, self.width,
                self.height, 0)
        if self.do_clear_buffer:
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    def render(self, alpha=1.0):
        if self.tex:
            glEnable(GL_TEXTURE_2D)
            glBindTexture(GL_TEXTURE_2D, self.tex)
            glColor4f(1.0, 1.0, 1.0, alpha)
            glBegin(GL_QUADS)
            glTexCoord2f(0.0, 0.0)
            glVertex2f(0.0, 0.0)
            glTexCoord2f(1.0, 0.0)
            glVertex2f(self.width, 0.0)
            glTexCoord2f(1.0, 1.0)
            glVertex2f(self.width, self.height)
            glTexCoord2f(0.0, 1.0)
            glVertex2f(0.0, self.height)
            glEnd()
