from __future__ import with_statement
from __future__ import absolute_import
from __future__ import division


from OpenGL.GL import *
from OpenGL.GLU import *

from OpenGL.GL.ARB.shader_objects import *
from OpenGL.GL.ARB.fragment_shader import *
from OpenGL.GL.ARB.vertex_shader import *


manager = None
has_shaders = None

def get_default_manager():
    global manager
    if not manager:
        manager = ShaderManager()
        return manager
    return manager

def does_support_shaders():
    global has_shaders
    if has_shaders is None:
        try:
            get_default_manager()
            has_shaders = True
        except Exception:
            has_shaders = False
    return has_shaders


class Program(object):
    def __init__(self, vertex_shader, fragment_shader):
        self.program, self.enable_func = get_default_manager().create_program(
                vertex_shader, fragment_shader)

    def set_uniform(self, name, value, type):
        try:
            manager._set_uniform(self.program, name, type, value)
        except:
            raise Exception("Uniform '%s' not supported!" % type)

    def enable(self):
        if self.enable_func:
            self.enable_func(self.program)
        else:
            raise Exception("No enable program func found")

    def disable(self):
        if self.enable_func:
            self.enable_func(0)
        else:
            raise Exception("No enable program func found")


class ShaderManager(object):
    def __init__(self):
        self.has_shaders = None
        self.create_program_func = None
        self.__find_shader_type()

    def gl_version(self):
        """Returns a list with the GL version numbers. I.e. [2, 1, 0]."""
        return [int(x) for x in glGetString(GL_VERSION).split()[0].split(".")]

    def __find_shader_type(self):
        if self.gl_version()[0] >= 2:
            print "We have shader support through opengl 2.0"
            self.has_opengl2 = True
            self.has_arb = False
            self.create_program_func = self.__create_shader_opengl2
        elif glCreateShaderObjectARB:
            print "We have shader support through ARB extensions"
            self.has_opengl2 = False
            self.has_arb = True
            self.create_program_func = self.__create_shader_arb
        else:
            self.has_arb = False
            self.has_opengl2 = False
            raise Exception("No shader support")

    def create_program(self, vertex_source, fragment_source):
        return self.create_program_func(vertex_source, fragment_source)

    def __create_shader_opengl2(self, vertex_source, fragment_source):
	if isinstance(vertex_source, (str,unicode)):
		vertex_source = [vertex_source]
	if isinstance(fragment_source, (str,unicode)):
		fragment_source = [fragment_source]

        # Create shaders
        v_shader = None
        if vertex_source:
            v_shader = glCreateShader(GL_VERTEX_SHADER)
            glShaderSource(v_shader, vertex_source)
            glCompileShader(v_shader)

        f_shader = None
        if fragment_source:
            f_shader = glCreateShader(GL_FRAGMENT_SHADER)
            glShaderSource(f_shader, fragment_source)
            glCompileShader(f_shader)

        return self.__create_program_opengl2(v_shader, f_shader)

    def __create_shader_arb(self, vertex_source, fragment_source):
	if isinstance(vertex_source, (str,unicode)):
		vertex_source = [vertex_source]
	if isinstance(fragment_source, (str,unicode)):
		fragment_source = [fragment_source]

        # Create shaders
        v_shader = None
        if vertex_source:
            v_shader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB)
            glShaderSourceARB(v_shader, vertex_source)
            glCompileShaderARB(v_shader)

        f_shader = None
        if fragment_source:
            f_shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB)
            glShaderSourceARB(f_shader, fragment_source)
            glCompileShaderARB(f_shader)

        return self.__create_program_arb(v_shader, f_shader)

    def __create_program_opengl2(self, vertex_shader, fragment_shader):
        program = glCreateProgram()

        # Attach the shaders
        if vertex_shader:
            glAttachShader(program, vertex_shader)
        if fragment_shader:
            glAttachShader(program, fragment_shader)

        # Link the program
        glLinkProgram(program)
        return program, self._enable_program_opengl2

    def __create_program_arb(self, vertex_shader, fragment_shader):
        program = glCreateProgramObjectARB()

        # Attach the shaders
        if vertex_shader:
            glAttachObjectARB(program, vertex_shader)
        if fragment_shader:
            glAttachObjectARB(program, fragment_shader)

        # Link the program
        glLinkProgramARB(program)
        return program, self._enable_program_arb

    def _enable_program_opengl2(self, program):
        glUseProgram(program)

    def _enable_program_arb(self, program):
        glUseProgramObjectARB(program)

    def _set_uniform_1i(self, program, name, value):
        if self.has_opengl2:
            loc = glGetUniformLocation(program, name)
            glUniform1i(loc, value)
        elif self.has_arb:
            loc = glGetUniformLocationARB(program, name)
            glUniform1iARB(loc, value)
        else:
            raise Exception("Cannot call this without shader support!")

    def _set_uniform_1f(self, program, name, value):
        if self.has_opengl2:
            loc = glGetUniformLocation(program, name)
            glUniform1f(loc, value)
        elif self.has_arb:
            loc = glGetUniformLocationARB(program, name)
            glUniform1fARB(loc, value)
        else:
            raise Exception("Cannot call this without shader support!")

    def _set_uniform(self, program, name, type, value):
        if self.has_opengl2:
            loc = glGetUniformLocation(program, name)
            if isinstance(value, (tuple, list)):
                globals()["glUniform" + type](loc, *value)
            else:
                globals()["glUniform" + type](loc, value)
        elif self.has_arb:
            loc = glGetUniformLocationARB(program, name)
            if isinstance(value, (tuple, list)):
                globals()["glUniform" + type + "ARB"](loc, *value)
            else:
                globals()["glUniform" + type + "ARB"](loc, value)
        else:
            raise Exception("Cannot call this without shader support!")

